@@ -16,9 +16,11 @@ Usage:
 On Windows, no additional modules are needed.
 On Mac, the pyobjc module is used, falling back to the pbcopy and pbpaste cli
     commands. (These commands should come with OS X.).
-On Linux, install xclip or xsel via package manager. For example, in Debian:
+On Linux, install xclip, xsel, or wl-clipboard (for "wayland" sessions) via package manager.
+For example, in Debian:
     sudo apt-get install xclip
     sudo apt-get install xsel
+    sudo apt-get install wl-clipboard
 
 Otherwise on Linux, you will need the gtk or PyQt5/PyQt4 modules installed.
 
@@ -32,18 +34,18 @@ Cygwin is currently not supported.
 
 Security Note: This module runs programs with these names:
     - which
-    - where
     - pbcopy
     - pbpaste
     - xclip
     - xsel
+    - wl-copy/wl-paste
     - klipper
     - qdbus
 A malicious user could rename or add programs with these names, tricking
 Pyperclip into running them with whatever permissions the Python process has.
 
 """
-__version__ = '1.6.5'
+__version__ = '1.9.0'
 
 import contextlib
 import ctypes
@@ -55,34 +57,32 @@ import time
 import warnings
 
 from ctypes import c_size_t, sizeof, c_wchar_p, get_errno, c_wchar
+from typing import Union, Optional
 
 
-# `import PyQt4` sys.exit()s if DISPLAY is not in the environment.
-# Thus, we need to detect the presence of $DISPLAY manually
-# and not load PyQt4 if it is absent.
-HAS_DISPLAY = os.getenv("DISPLAY", False)
+_IS_RUNNING_PYTHON_2 = sys.version_info[0] == 2  # type: bool
 
-EXCEPT_MSG = """
-    Pyperclip could not find a copy/paste mechanism for your system.
-    For more information, please visit https://pyperclip.readthedocs.io/en/latest/introduction.html#not-implemented-error """
-
-PY2 = sys.version_info[0] == 2
-
-STR_OR_UNICODE = unicode if PY2 else str # For paste(): Python 3 uses str, Python 2 uses unicode.
-
-ENCODING = 'utf-8'
-
-# The "which" unix command finds where a command is.
-if platform.system() == 'Windows':
-    WHICH_CMD = 'where'
+# For paste(): Python 3 uses str, Python 2 uses unicode.
+if _IS_RUNNING_PYTHON_2:
+    # mypy complains about `unicode` for Python 2, so we ignore the type error:
+    _PYTHON_STR_TYPE = unicode  # type: ignore
 else:
-    WHICH_CMD = 'which'
-
-def _executable_exists(name):
-    return subprocess.call([WHICH_CMD, name],
-                           stdout=subprocess.PIPE, stderr=subprocess.PIPE) == 0
+    _PYTHON_STR_TYPE = str
 
+ENCODING = 'utf-8'  # type: str
 
+try:
+    # Use shutil.which() for Python 3+
+    from shutil import which
+    def _py3_executable_exists(name):  # type: (str) -> bool
+        return bool(which(name))
+    _executable_exists = _py3_executable_exists
+except ImportError:
+    # Use the "which" unix command for Python 2.7 and prior.
+    def _py2_executable_exists(name):  # type: (str) -> bool
+        return subprocess.call(['which', name],
+                            stdout=subprocess.PIPE, stderr=subprocess.PIPE) == 0
+    _executable_exists = _py2_executable_exists
 
 # Exceptions
 class PyperclipException(RuntimeError):
@@ -93,21 +93,13 @@ class PyperclipWindowsException(PyperclipException):
         message += " (%s)" % ctypes.WinError()
         super(PyperclipWindowsException, self).__init__(message)
 
-
-def _stringifyText(text):
-    if PY2:
-        acceptedTypes = (unicode, str, int, float, bool)
-    else:
-        acceptedTypes = (str, int, float, bool)
-    if not isinstance(text, acceptedTypes):
-        raise PyperclipException('only str, int, float, and bool values can be copied to the clipboard, not %s' % (text.__class__.__name__))
-    return STR_OR_UNICODE(text)
+class PyperclipTimeoutException(PyperclipException):
+    pass
 
 
 def init_osx_pbcopy_clipboard():
-
     def copy_osx_pbcopy(text):
-        text = _stringifyText(text) # Converts non-str values to str.
+        text = _PYTHON_STR_TYPE(text) # Converts non-str values to str.
         p = subprocess.Popen(['pbcopy', 'w'],
                              stdin=subprocess.PIPE, close_fds=True)
         p.communicate(input=text.encode(ENCODING))
@@ -124,7 +116,7 @@ def init_osx_pbcopy_clipboard():
 def init_osx_pyobjc_clipboard():
     def copy_osx_pyobjc(text):
         '''Copy string argument to clipboard'''
-        text = _stringifyText(text) # Converts non-str values to str.
+        text = _PYTHON_STR_TYPE(text) # Converts non-str values to str.
         newStr = Foundation.NSString.stringWithString_(text).nsstring()
         newData = newStr.dataUsingEncoding_(Foundation.NSUTF8StringEncoding)
         board = AppKit.NSPasteboard.generalPasteboard()
@@ -146,7 +138,7 @@ def init_gtk_clipboard():
 
     def copy_gtk(text):
         global cb
-        text = _stringifyText(text) # Converts non-str values to str.
+        text = _PYTHON_STR_TYPE(text) # Converts non-str values to str.
         cb = gtk.Clipboard()
         cb.set_text(text)
         cb.store()
@@ -180,13 +172,13 @@ def init_qt_clipboard():
         app = QApplication([])
 
     def copy_qt(text):
-        text = _stringifyText(text) # Converts non-str values to str.
+        text = _PYTHON_STR_TYPE(text) # Converts non-str values to str.
         cb = app.clipboard()
         cb.setText(text)
 
     def paste_qt():
         cb = app.clipboard()
-        return STR_OR_UNICODE(cb.text())
+        return _PYTHON_STR_TYPE(cb.text())
 
     return copy_qt, paste_qt
 
@@ -196,7 +188,7 @@ def init_xclip_clipboard():
     PRIMARY_SELECTION='p'
 
     def copy_xclip(text, primary=False):
-        text = _stringifyText(text) # Converts non-str values to str.
+        text = _PYTHON_STR_TYPE(text) # Converts non-str values to str.
         selection=DEFAULT_SELECTION
         if primary:
             selection=PRIMARY_SELECTION
@@ -224,7 +216,7 @@ def init_xsel_clipboard():
     PRIMARY_SELECTION='-p'
 
     def copy_xsel(text, primary=False):
-        text = _stringifyText(text) # Converts non-str values to str.
+        text = _PYTHON_STR_TYPE(text) # Converts non-str values to str.
         selection_flag = DEFAULT_SELECTION
         if primary:
             selection_flag = PRIMARY_SELECTION
@@ -244,9 +236,36 @@ def init_xsel_clipboard():
     return copy_xsel, paste_xsel
 
 
+def init_wl_clipboard():
+    PRIMARY_SELECTION = "-p"
+
+    def copy_wl(text, primary=False):
+        text = _PYTHON_STR_TYPE(text)  # Converts non-str values to str.
+        args = ["wl-copy"]
+        if primary:
+            args.append(PRIMARY_SELECTION)
+        if not text:
+            args.append('--clear')
+            subprocess.check_call(args, close_fds=True)
+        else:
+            pass
+            p = subprocess.Popen(args, stdin=subprocess.PIPE, close_fds=True)
+            p.communicate(input=text.encode(ENCODING))
+
+    def paste_wl(primary=False):
+        args = ["wl-paste", "-n"]
+        if primary:
+            args.append(PRIMARY_SELECTION)
+        p = subprocess.Popen(args, stdout=subprocess.PIPE, close_fds=True)
+        stdout, _stderr = p.communicate()
+        return stdout.decode(ENCODING)
+
+    return copy_wl, paste_wl
+
+
 def init_klipper_clipboard():
     def copy_klipper(text):
-        text = _stringifyText(text) # Converts non-str values to str.
+        text = _PYTHON_STR_TYPE(text) # Converts non-str values to str.
         p = subprocess.Popen(
             ['qdbus', 'org.kde.klipper', '/klipper', 'setClipboardContents',
              text.encode(ENCODING)],
@@ -275,7 +294,7 @@ def init_klipper_clipboard():
 
 def init_dev_clipboard_clipboard():
     def copy_dev_clipboard(text):
-        text = _stringifyText(text) # Converts non-str values to str.
+        text = _PYTHON_STR_TYPE(text) # Converts non-str values to str.
         if text == '':
             warnings.warn('Pyperclip cannot copy a blank string to the clipboard on Cygwin. This is effectively a no-op.')
         if '\r' in text:
@@ -298,9 +317,12 @@ def init_no_clipboard():
     class ClipboardUnavailable(object):
 
         def __call__(self, *args, **kwargs):
-            raise PyperclipException(EXCEPT_MSG)
+            additionalInfo = ''
+            if sys.platform == 'linux':
+                additionalInfo = '\nOn Linux, you can run `sudo apt-get install xclip` or `sudo apt-get install xselect` to install a copy/paste mechanism.'
+            raise PyperclipException('Pyperclip could not find a copy/paste mechanism for your system. For more information, please visit https://pyperclip.readthedocs.io/en/latest/index.html#not-implemented-error' + additionalInfo)
 
-        if PY2:
+        if _IS_RUNNING_PYTHON_2:
             def __nonzero__(self):
                 return False
         else:
@@ -425,7 +447,7 @@ def init_windows_clipboard():
         # This function is heavily based on
         # http://msdn.com/ms649016#_win32_Copying_Information_to_the_Clipboard
 
-        text = _stringifyText(text) # Converts non-str values to str.
+        text = _PYTHON_STR_TYPE(text) # Converts non-str values to str.
 
         with window() as hwnd:
             # http://msdn.com/ms649048
@@ -460,20 +482,24 @@ def init_windows_clipboard():
                 # (Also, it may return a handle to an empty buffer,
                 # but technically that's not empty)
                 return ""
-            return c_wchar_p(handle).value
+            locked_handle = safeGlobalLock(handle)
+            return_value = c_wchar_p(locked_handle).value
+            safeGlobalUnlock(handle)
+            return return_value
 
     return copy_windows, paste_windows
 
 
 def init_wsl_clipboard():
     def copy_wsl(text):
-        text = _stringifyText(text) # Converts non-str values to str.
+        text = _PYTHON_STR_TYPE(text) # Converts non-str values to str.
         p = subprocess.Popen(['clip.exe'],
                              stdin=subprocess.PIPE, close_fds=True)
         p.communicate(input=text.encode(ENCODING))
 
     def paste_wsl():
-        p = subprocess.Popen(['powershell.exe', '-command', 'Get-Clipboard'],
+        # '-noprofile' speeds up load time
+        p = subprocess.Popen(['powershell.exe', '-noprofile', '-command', 'Get-Clipboard'],
                              stdout=subprocess.PIPE,
                              stderr=subprocess.PIPE,
                              close_fds=True)
@@ -505,9 +531,9 @@ def determine_clipboard():
     elif os.name == 'nt' or platform.system() == 'Windows':
         return init_windows_clipboard()
 
-    if platform.system() == 'Linux':
+    if platform.system() == 'Linux' and os.path.isfile('/proc/version'):
         with open('/proc/version', 'r') as f:
-            if "Microsoft" in f.read():
+            if "microsoft" in f.read().lower():
                 return init_wsl_clipboard()
 
     # Setup for the MAC OS X platform:
@@ -521,7 +547,11 @@ def determine_clipboard():
             return init_osx_pyobjc_clipboard()
 
     # Setup for the LINUX platform:
-    if HAS_DISPLAY:
+
+    # `import PyQt4` sys.exit()s if DISPLAY is not in the environment.
+    # Thus, we need to detect the presence of $DISPLAY manually
+    # and not load PyQt4 if it is absent.
+    if os.getenv("DISPLAY"):
         try:
             import gtk  # check if gtk is installed
         except ImportError:
@@ -529,10 +559,16 @@ def determine_clipboard():
         else:
             return init_gtk_clipboard()
 
+        if (
+                os.getenv("WAYLAND_DISPLAY") and
+                _executable_exists("wl-copy")
+        ):
+            return init_wl_clipboard()
+        if _executable_exists("xclip"):
+            # Note: 2024/06/18 Google Trends shows xclip as more popular than xsel.
+            return init_xclip_clipboard()
         if _executable_exists("xsel"):
             return init_xsel_clipboard()
-        if _executable_exists("xclip"):
-            return init_xclip_clipboard()
         if _executable_exists("klipper") and _executable_exists("qdbus"):
             return init_klipper_clipboard()
 
@@ -577,15 +613,18 @@ def set_clipboard(clipboard):
     '''
     global copy, paste
 
-    clipboard_types = {'pbcopy': init_osx_pbcopy_clipboard,
-                       'pyobjc': init_osx_pyobjc_clipboard,
-                       'gtk': init_gtk_clipboard,
-                       'qt': init_qt_clipboard, # TODO - split this into 'qtpy', 'pyqt4', and 'pyqt5'
-                       'xclip': init_xclip_clipboard,
-                       'xsel': init_xsel_clipboard,
-                       'klipper': init_klipper_clipboard,
-                       'windows': init_windows_clipboard,
-                       'no': init_no_clipboard}
+    clipboard_types = {
+        "pbcopy": init_osx_pbcopy_clipboard,
+        "pyobjc": init_osx_pyobjc_clipboard,
+        "gtk": init_gtk_clipboard,
+        "qt": init_qt_clipboard,  # TODO - split this into 'qtpy', 'pyqt4', and 'pyqt5'
+        "xclip": init_xclip_clipboard,
+        "xsel": init_xsel_clipboard,
+        "wl-clipboard": init_wl_clipboard,
+        "klipper": init_klipper_clipboard,
+        "windows": init_windows_clipboard,
+        "no": init_no_clipboard,
+    }
 
     if clipboard not in clipboard_types:
         raise ValueError('Argument must be one of %s' % (', '.join([repr(_) for _ in clipboard_types.keys()])))
@@ -648,6 +687,7 @@ def is_available():
 copy, paste = lazy_load_stub_copy, lazy_load_stub_paste
 
 
+
 __all__ = ['copy', 'paste', 'set_clipboard', 'determine_clipboard']
 
 
